#include <cassert>
#include <cstring>
#include "Lexer.h"

Lexer::Lexer(int maxpatterns) : mpattern(new ref_ptr<Pattern>[maxpatterns]), mpatterns(0) {
}

Lexer::~Lexer() {
	delete [] mpattern;
}

void Lexer::add(int index, char const *, bool ignore = false, bool enable = true) {
	addPattern(index, new Pattern(ignore, enable));
}

void Lexer::addPattern(int index, Pattern *pattern) {
	assert(index > 255);
	index -= 256;
	mpattern[index] = pattern;
	if (index == mpatterns) ++mpatterns;
}

Lexer::MatchResult Lexer::get(string &out, char const *&msource, char const *&mmatch, int &mtype) {
bool ignore;

	do {
		ignore = false;
		for (int i = 0; i < mpatterns; ++i) {
		Pattern *p = mpattern[i];
		int size;

			if (p->enabled)
				if ((size = p->match(msource)) != -1) {
					if (p->ignored) {
						msource += size;
						ignore = true;
						if (*msource == 0) {
							msource = 0;
							return MATCH_NONE;
						}
						break;
					} else {
						mmatch = msource;
						msource += size;
						mtype = i + 256;
						return MATCH_SOURCE;
					}
				}
		}
	} while (*msource && ignore);
	return MATCH_NONE;
}

bool Lexer::preget(string &out, char const *&msource, char const *&mmatch, int &mtype, int &msize, int &mline) {
	if (msource) {
		if (*msource == 0) {
			msource = 0;
			mmatch = 0;
			msize = -1;
			mtype = -1;
			out = "";
		} else {
		char const *mark = msource;
		MatchResult result;
			if ((result = get(out, msource, mmatch, mtype)) != MATCH_NONE) {
				while (mark != msource) {
					if (*mark == '\n') ++mline;
					++mark;
				}
				msize = msource - mmatch;
				if (result == MATCH_SOURCE) out = string(mmatch, msize);
				return true;
			} else {
				if (msource && *msource) {
					mmatch = msource;
					mtype = *msource;
					msource++;
					while (mark != msource) {
						if (*mark == '\n') ++mline;
						++mark;
					}
					out = string(msource - 1, 1);
					msize = 1;
					return true;
				} else {
					msize = -1;
					msource = mmatch = 0;
					mtype = -1;
					out = "";
				}
			}
		}
	}
	return false;
}
